import functools
import importlib
import weakref

try:
    unicode_str = unicode
except NameError:
    unicode_str = str


def metaclass(mcs):
    def _decorator(cls):
        attrs = dict(vars(cls))
        try:
            if isinstance(cls.__slots__, str):
                slots = (cls.__slots__, )
            else:
                slots = cls.__slots__
            for slot in slots:
                del attrs[slot]
        except AttributeError:
            pass
        for prop in '__weakref__', '__dict__':
            attrs.pop(prop, None)
        return mcs(cls.__name__, cls.__bases__, attrs)
    return _decorator


class CachedDescriptor(object):

    __slots__ = '_cache',

    def __init__(self):
        self._cache = weakref.WeakKeyDictionary()

    def __get__(self, instance, instance_type=None):
        if instance is None:
            return self  # static access
        try:
            return self._cache[instance]
        except KeyError:
            value = self._cache[instance] = self.get_value(instance)
            return value

    def __set__(self, instance, value):
        self._cache[instance] = self.set_value(instance, value)

    def __delete__(self, instance):
        del self._cache[instance]

    def get_value(self, instance):
        raise NotImplementedError

    def set_value(self, instance, value):
        return value


class cached_property(CachedDescriptor):

    __slots__ = 'get', 'set', '__doc__'

    def __init__(self, fget=None, fset=None, doc=None):
        super(cached_property, self).__init__()
        self.get = fget
        self.set = fset
        self.__doc__ = doc

    def get_value(self, instance):
        if self.get is None:
            raise AttributeError("unreadable attribute")
        return self.get(instance)

    def set_value(self, instance, value):
        if self.set is not None:
            return self.set(instance, value)
        return super(cached_property, self).set_value(instance, value)

    def getter(self, getter):
        self.get = getter
        return self

    def setter(self, setter):
        self.set = setter
        return self


class LazyType(type):

    def __call__(cls, *args):
        assert len(args) < 2
        try:
            arg = args[0]
            if isinstance(arg, cls) or not isinstance(arg, str):
                return arg
        except IndexError:
            pass
        return super(LazyType, cls).__call__(*args)


@metaclass(LazyType)
class Lazy(CachedDescriptor):

    __slots__ = '_path', '__weakref__'

    def __init__(self, path=None):
        super(Lazy, self).__init__()
        self._path = path

    def __call__(self, *args, **kwargs):
        return self._obj(*args, **kwargs)

    def __iter__(self):
        return iter(self._obj)

    def __len__(self):
        return len(self._obj)

    def __str__(self):
        return str(self._obj)

    def __unicode__(self):
        return unicode(self._obj)

    def __bytes__(self):
        return bytes(self._obj)

    def __getitem__(self, item):
        return self._obj[item]

    def __getattr__(self, attr):
        return getattr(self._obj, attr)

    @property
    def __class__(self):
        return self._obj.__class__

    def __get__(self, instance, instance_type=None):
        value = super(Lazy, self).__get__(instance, instance_type)
        return value._obj if isinstance(value, Lazy) else value

    @cached_property
    def _obj(self):
        try:
            module_name, obj_name = self._path.rsplit('.', 1)
        except ValueError:
            module_name, obj_name = self._path, None
        module = importlib.import_module(module_name)
        if obj_name is not None:
            return getattr(module, obj_name)
        return module

    @classmethod
    def set_value(cls, instance, value):
        return cls(value)


def to_bytes(obj, encoding='utf-8', error_callback=None):
    try:
        if isinstance(obj, (bytes, bytearray, memoryview)):
            return bytes(obj)
        if obj is None:
            return b''
        try:
            return obj.__bytes__()
        except AttributeError:
            return unicode_str(obj).encode(encoding=encoding)
    except Exception as error:
        if error_callback is not None:
            error_callback(error)
        raise


def coroutine(fn):
    @functools.wraps(fn)
    def _fn(*args, **kwargs):
        co = fn(*args, **kwargs)
        co.send(None)
        return co
    return _fn


def is_class(cls):
    return issubclass(type(cls), type)



import collections
import functools
import itertools
import logging

from marnadi import descriptors, Header
from marnadi.errors import HttpError
from marnadi.utils import to_bytes, cached_property, coroutine

try:
    str = unicode
except NameError:
    pass


class Response(collections.Iterator):

    logger = logging.getLogger('marnadi')

    status = '200 OK'

    supported_http_methods = {
        'OPTIONS', 'GET', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE',
    }

    if hasattr(collections.Iterator, '__slots__'):
        __slots__ = 'app', 'request', '__weakref__'

    headers = descriptors.Headers(
        ('Content-Type', Header('text/plain', charset='utf-8')),
    )

    cookies = descriptors.Cookies()

    def __init__(self, app, request):
        self.app = app
        self.request = request

    def __call__(self, **kwargs):
        if self.request.method not in self.supported_http_methods:
            raise HttpError(
                '501 Not Implemented',
                headers=(('Allow', ', '.join(self.allowed_http_methods)), )
            )
        callback = getattr(self, self.request.method.lower())
        if callback is None:
            raise HttpError(
                '405 Method Not Allowed',
                headers=(('Allow', ', '.join(self.allowed_http_methods)), )
            )
        return callback(**kwargs)

    @classmethod
    def __subclasshook__(cls, subclass):
        return isinstance(subclass, cls.FunctionHandler) or NotImplemented

    def __next__(self):
        return next(self.iterator)

    def next(self):
        return self.__next__()

    @cached_property
    @coroutine
    def iterator(self):
        result = yield
        if result is None or isinstance(result, (str, bytes)):
            chunk = to_bytes(result)
            self.headers.setdefault('Content-Length', len(chunk))
            yield chunk
        else:
            chunks = iter(result)
            first_chunk = to_bytes(next(chunks, b''))
            try:
                result_length = len(result)
            except TypeError:  # result doesn't support len()
                pass
            else:
                if result_length <= 1:
                    self.headers.setdefault(
                        'Content-Length',
                        len(first_chunk),
                    )
            yield first_chunk
            for chunk in chunks:
                yield to_bytes(chunk, error_callback=self.logger.exception)

    @classmethod
    def start(cls, *args, **params):
        try:
            response = cls(*args)
            result = response(**params)
            response.iterator = itertools.chain(
                (response.iterator.send(result), ),
                response.iterator
            )
            return response
        except HttpError:
            raise
        except Exception as error:
            cls.logger.exception(error)
            raise

    class FunctionHandler(type):

        __function__ = NotImplemented

        __response__ = NotImplemented

        def __call__(cls, *args, **kwargs):
            return cls.__function__(*args, **kwargs)

        def start(cls, **kwargs):
            raise NotImplementedError

        class classmethod(classmethod):

            def __get__(self, instance, cls):
                assert isinstance(cls, Response.FunctionHandler)
                return functools.partial(self.__func__, cls.__response__)

    @classmethod
    def handler(cls, *methods):
        def decorator(func):
            method = staticmethod(func)
            attributes = dict(
                __module__=func.__module__,
                __doc__=func.__doc__,
                __slots__=(),
            )
            response_class = type(func.__name__, (cls, ), dict(
                {m.lower(): method for m in methods},
                **attributes
            ))
            func_replacement = cls.FunctionHandler(
                func.__name__,
                (),
                dict(
                    attributes,
                    __function__=method,
                    __response__=response_class,
                    start=cls.FunctionHandler.classmethod(cls.start.__func__),
                ),
            )
            return func_replacement
        return decorator

    @property
    def allowed_http_methods(self):
        for method in self.supported_http_methods:
            if getattr(self, method.lower()):
                yield method

    def options(self, **kwargs):
        self.headers['Allow'] = ', '.join(self.allowed_http_methods)

    get = None

    head = None

    post = None

    put = None

    patch = None

    delete = None
